"""
web2ldap plugin classes for Active Directory (for some information see draft-armijo-ldap-syntax)

$Id: activedirectory.py,v 1.13 2009/03/29 11:19:32 michael Exp $
"""

import time,utctime,struct,pyweblib,ldap.cidict,w2lapp.cnf,mspki.util

from w2lapp.gui import DisplayDN
from w2lapp.core import utf2display

from w2lapp.schema.syntaxes import \
  LDAPSyntax,Binary,Boolean,DistinguishedName,DirectoryString,GeneralizedTime, \
  Integer,BitArrayInteger,OctetString,SelectList,Uri,syntax_registry


class ObjectSID(OctetString):
  oid = 'ObjectSID-oid'
  desc = 'Base class for Windows Security Identifiers'
  """
  SID anatomy:
  Byte Position
  0 : SID Structure Revision Level (SRL)
  1 : Number of Subauthority/Relative Identifier
  2-7 : Identifier Authority Value (IAV) [48 bits]
  8-x : Variable number of Subauthority or Relative Identifier (RID)
        [32 bits]
  """

  def _sid2sddl(self,sid):
    srl = ord(sid[0])
    number_sub_id = ord(sid[1])
    iav = struct.unpack('!Q','\x00\x00'+sid[2:8])[0]
    sub_ids = [
      struct.unpack('L',sid[8+4*i:12+4*i])[0]
      for i in range(number_sub_id)
    ]
    return 'S-%d-%d-%s' % (
      srl,
      iav,
      '-'.join([str(s) for s in sub_ids]),
    )

  def _sddl2sid(self,sddl):
    sid_components = sddl.split('-')
    srl_byte = chr(int(sid_components[1]))
    number_sub_id_byte = chr(len(sid_components)-3)
    iav_buf = struct.pack('!Q',int(sid_components[2]))[2:]
    result_list = [srl_byte,number_sub_id_byte,iav_buf]
    result_list.extend([
      struct.pack('L',int(s))
      for s in sid_components[3:]
    ])
    return ''.join(result_list)

  def displayValue(self,valueindex=0,commandbutton=0):
    sddl = self._sid2sddl(self.attrValue)
    return '%s<br>%s' % (
      self._sid2sddl(self.attrValue),
      OctetString.displayValue(self,valueindex,commandbutton)
    )


# Register certain attribute types for syntax classes
syntax_registry.registerAttrType(
  ObjectSID.oid,[
    '1.2.840.113556.1.4.146', # objectSID
    '1.2.840.113556.1.4.609', # sIDHistory
  ]
)


class OtherSID(ObjectSID):
  oid = 'OtherSID-oid'
  desc = 'SID in MS AD which points to another object'
  well_known_sids = {
    # see http://msdn.microsoft.com/en-us/library/aa379649(VS.85).aspx
    'S-1-2-32-544':'BUILTIN\ADMINISTRATORS',
    'S-1-2-32-545':'BUILTIN\USERS',
    'S-1-2-32-546':'BUILTIN\GUESTS',
    'S-1-2-32-548':'BUILTIN\ACCOUNT OPERATORS',
    'S-1-2-32-549':'BUILTIN\SERVER OPERATORS',
    'S-1-2-32-550':'BUILTIN\PRINT OPERATORS',
    'S-1-2-32-551':'BUILTIN\BACKUP OPERATORS',
    'S-1-2-32-552':'BUILTIN\REPLICATOR',
    'S-1-0-0':     'Null SID',
    'S-1-1-0':     'World',
    'S-1-2-0':     'Local',
    'S-1-3-0':     'Creator Owner ID',
    'S-1-3-1':     'Creator Group ID',
  }

  def displayValue(self,valueindex=0,commandbutton=0):
    sddl_str = self._sid2sddl(self.attrValue)
    search_anchor = self.well_known_sids.get(sddl_str,'')
    if commandbutton and not sddl_str in self.well_known_sids:
        search_anchor = self._form.applAnchor(
            'searchform','&raquo;',self._sid,
            [
              ('dn',self._dn),
              ('search_filterstr', '(objectSid=%s)' % (
                  utf2display(self._form.accept_charset,sddl_str,escape_entities=1)
                )
              ),
              ('searchform_mode','exp'),
            ],
            title='Search by SID',
        )
    return '%s %s<br>%s' % (
      sddl_str,search_anchor,
      OctetString.displayValue(self,valueindex,commandbutton),
    )

# Register certain attribute types for syntax classes
syntax_registry.registerAttrType(
  OtherSID.oid,[
    '1.2.840.113556.1.4.1301', # tokenGroups
    '1.2.840.113556.1.4.1418', # tokenGroupsGlobalAndUniversal
    '1.2.840.113556.1.4.1303', # tokenGroupsNoGCAcceptable
    '1.2.840.113556.1.4.667',  # syncWithSID
    '1.2.840.113556.1.4.1410', # mS-DS-CreatorSID
  ]
)


class SAMAccountName(DirectoryString):
  oid = 'SAMAccountName-oid'
  desc = 'SAM-Account-Name in MS AD'
  maxLen = 20

# Register certain attribute types for syntax classes
syntax_registry.registerAttrType(
  SAMAccountName.oid,[
    '1.2.840.113556.1.4.221', # sAMAccountName
  ]
)


class SAMAccountType(SelectList):
  """
  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/adschema/adschema/a_samaccounttype.asp
  """
  oid = 'SAMAccountType-oid'
  desc = 'SAM-Account-Type in MS AD'
  attr_value_dict = {
     '':'',
     '268435456':'SAM_GROUP_OBJECT',
     '268435457':'SAM_NON_SECURITY_GROUP_OBJECT',
     '536870912':'SAM_ALIAS_OBJECT',
     '536870913':'SAM_NON_SECURITY_ALIAS_OBJECT',
     '805306368':'SAM_NORMAL_USER_ACCOUNT',
     '805306369':'SAM_MACHINE_ACCOUNT',
     '805306370':'SAM_TRUST_ACCOUNT',
    '1073741824':'SAM_APP_BASIC_GROUP',
    '1073741825':'SAM_APP_QUERY_GROUP',
    '2147483647':'SAM_ACCOUNT_TYPE_MAX',
  }

# Register certain attribute types for syntax classes
syntax_registry.registerAttrType(
  SAMAccountType.oid,[
    '1.2.840.113556.1.4.302', # sAMAccountType
  ]
)


class GroupType(SelectList):
  """
  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/adschema/adschema/a_samaccounttype.asp
  """
  oid = 'GroupType-oid'
  desc = 'Group-Type in MS AD'
  attr_value_dict = {
     '':'',
    '-2147483640':'Universal / Security',
    '-2147483644':'Domain local / Security',
    '-2147483646':'Global / Security',
    '2':'Global / Distribution',
    '4':'Domain local / Distribution',
    '8':'Universal / Distribution',
  }

# Register certain attribute types for syntax classes
syntax_registry.registerAttrType(
  GroupType.oid,[
    '1.2.840.113556.1.4.750', # groupType
  ]
)


class UserAccountControl(BitArrayInteger):
  """
  See knowledge base article 305144:
  http://support.microsoft.com/default.aspx?scid=kb;en-us;Q305144
  """
  oid = 'UserAccountControl-oid'
  flag_desc_table = (
    ('SCRIPT',0x0001),
    ('ACCOUNTDISABLE',0x0002),
    ('HOMEDIR_REQUIRED',0x0008),
    ('LOCKOUT',0x0010),
    ('PASSWD_NOTREQD',0x0020),
    ('PASSWD_CANT_CHANGE',0x0040),
    ('ENCRYPTED_TEXT_PWD_ALLOWED',0x0080),
    ('TEMP_DUPLICATE_ACCOUNT',0x0100),
    ('NORMAL_ACCOUNT',0x0200),
    ('INTERDOMAIN_TRUST_ACCOUNT',0x0800),
    ('WORKSTATION_TRUST_ACCOUNT',0x1000),
    ('SERVER_TRUST_ACCOUNT',0x2000),
    ('DONT_EXPIRE_PASSWORD',0x10000),
    ('MNS_LOGON_ACCOUNT',0x20000),
    ('SMARTCARD_REQUIRED',0x40000),
    ('TRUSTED_FOR_DELEGATION',0x80000),
    ('NOT_DELEGATED',0x100000),
    ('USE_DES_KEY_ONLY',0x200000),
    ('DONT_REQ_PREAUTH',0x400000),
    ('PASSWORD_EXPIRED',0x800000),
    ('TRUSTED_TO_AUTH_FOR_DELEGATION',0x1000000)
  )

syntax_registry.registerAttrType(
  UserAccountControl.oid,[
    '1.2.840.113556.1.4.8', # userAccountControl
  ]
)


class SystemFlags(BitArrayInteger):
  """
  See
  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/adschema/adschema/a_systemflags.asp
  and
  http://msdn2.microsoft.com/en-us/library/aa772297.aspx
  """
  oid = 'SystemFlags-oid'
  flag_desc_table = (
    ('ADS_SYSTEMFLAG_DISALLOW_DELETE',0x80000000),
    ('ADS_SYSTEMFLAG_CONFIG_ALLOW_RENAME',0x40000000),
    ('ADS_SYSTEMFLAG_CONFIG_ALLOW_MOVE',0x20000000),
    ('ADS_SYSTEMFLAG_CONFIG_ALLOW_LIMITED_MOVE',0x10000000),
    ('ADS_SYSTEMFLAG_DOMAIN_DISALLOW_RENAME',0x08000000),
    ('ADS_SYSTEMFLAG_DOMAIN_DISALLOW_MOVE',0x04000000),
    ('ADS_SYSTEMFLAG_CR_NTDS_NC',0x00000001),
    ('ADS_SYSTEMFLAG_CR_NTDS_DOMAIN',0x00000002),
    ('ADS_SYSTEMFLAG_ATTR_NOT_REPLICATED',0x00000001),
    ('ADS_SYSTEMFLAG_ATTR_IS_CONSTRUCTED',0x00000004),
  )

syntax_registry.registerAttrType(
  SystemFlags.oid,[
    '1.2.840.113556.1.4.375', # systemFlags
  ]
)


class SearchFlags(BitArrayInteger):
  """
  http://msdn.microsoft.com/en-us/library/ms679765(VS.85).aspx

   1 (0x00000001)   Create an index for the attribute.
   2 (0x00000002)   Create an index for the attribute in each container.
   4 (0x00000004)   Add this attribute to the Ambiguous Name Resolution (ANR) set. This is used to assist in finding an object when only partial information is given. For example, if the LDAP filter is (ANR=JEFF), the search will find each object where the first name, last name, e-mail address, or other ANR attribute is equal to JEFF. Bit 0 must be set for this index take affect.
   8 (0x00000008)   Preserve this attribute in the tombstone object for deleted objects.
  16 (0x00000010)   Copy the value for this attribute when the object is copied.
  32 (0x00000020)   Supported beginning with Windows Server 2003. Create a tuple index for the attribute. This will improve searches where the wildcard appears at the front of the search string. For example, (sn=*mith).
  64 (0x00000040)   Supported beginning with ADAM. Creates an index to greatly help VLV performance on arbitrary attributes.
  """
  oid = 'SearchFlags-oid'
  flag_desc_table = (
    ('Indexed',0x0001),
    ('Indexed in each container',0x0002),
    ('Ambiguous Name Resolution (ANR)',0x0004),
    ('Preserve in tombstone object',0x0008),
    ('Copy value when object copied',0x0010),
    ('tuple index',0x0020),
    ('VLV index (Subtree Index in ADAM)',0x0040),
    ('CONFIDENTIAL',0x0080),
    ('NEVER_AUDIT_VALUE',0x0100),
    ('RODC_FILTERED',0x0200),
    ('',0x0400),
    ('',0x0800),
  )

syntax_registry.registerAttrType(
  SearchFlags.oid,[
    '1.2.840.113556.1.2.334', # searchFlags
  ]
)


class LogonHours(OctetString):
  oid = 'LogonHours-oid'
  desc = 'Logon hours'
  dayofweek = ('Sun','Mon','Tue','Wed','Thu','Fri','Sat')

  def _extractHours(self,value):
    if not value or len(value)!=21:
      return []
    hour_flags = []
    for eight_hours in value:
      eight_hours_int = ord(eight_hours)
      for i in range(8):
        hour_flags.append({0:'-',1:'X'}[(eight_hours_int>>i)&1])
    # For whatever reason the list has to be shifted one hour
    return hour_flags

  def sanitizeInput(self,inputValue):
    if not inputValue:
      return ''
    inputValue = inputValue.replace('\r','').replace('\n','')
    hour_flags = [
      int(inputValue[i]=='X')<<i%8
      for i in xrange(len(inputValue))
    ]
    r = [
      chr(sum(hour_flags[i*8:(i+1)*8]))
      for i in xrange(21)
    ]
    attrValue = ''.join(r)
    return attrValue

  def validate(self,attrValue):
    return len(attrValue)==21

  def formField(self):
    hour_flags = self._extractHours(self.attrValue)
    if hour_flags:
      day_bits = [
        ''.join(hour_flags[24*day:24*day+24])
        for day in range(7)
      ]
    else:
      day_bits = []
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,1,
      None,
      default=u'\r\n'.join(day_bits),
      rows=7,cols=24
    )

  def displayValue(self,valueindex=0,commandbutton=0):
    hour_flags = self._extractHours(self.attrValue)
    result_lines = [
      """<tr>
      <th width="10%%">Day</th>
      <th colspan="3" width="8%%">0</th>
      <th colspan="3" width="8%%">3</th>
      <th colspan="3" width="8%%">6</th>
      <th colspan="3" width="8%%">9</th>
      <th colspan="3" width="8%%">12</th>
      <th colspan="3" width="8%%">15</th>
      <th colspan="3" width="8%%">18</th>
      <th colspan="3" width="8%%">21</th>
    </tr>""",
    ]
    for day in range(7):
      day_bits = hour_flags[24*day:24*day+24]
      result_lines.append(
        '<tr><td>%s</td><td>%s</td></tr>' % (
          self.dayofweek[day],
          '</td><td>'.join(day_bits)
      ))
    return '<p>%s</p><table border>%s</table>' % (
      OctetString.displayValue(self,valueindex,commandbutton),
      '\n'.join(result_lines)
    )


syntax_registry.registerAttrType(
  LogonHours.oid,[
    '1.2.840.113556.1.4.64', # logonHours
  ]
)


class CountryCode(SelectList):
  oid = 'CountryCode-oid'
  desc = 'Numerical country code'
  attr_value_dict = ldap.cidict.cidict(w2lapp.cnf.countries.countryCode_dict)


syntax_registry.registerAttrType(
  CountryCode.oid,[
    '1.2.840.113556.1.4.25', # countryCode
  ]
)


class InstanceType(BitArrayInteger):
  """
  http://msdn.microsoft.com/library/en-us/adschema/adschema/a_instancetype.asp
  """
  oid = 'InstanceType-oid'
  flag_desc_table = (
    ('The head of naming context.',0x00000001),
    ('This replica is not instantiated.',0x00000002),
    ('The object is writable on this directory.',0x00000004),
    ('The naming context above this one on this directory is held.',0x00000008),
    ('The naming context is in the process of being constructed for the first time via replication.',0x00000010),
    ('The naming context is in the process of being removed from the local DSA.',0x00000020),
  )

syntax_registry.registerAttrType(
  InstanceType.oid,[
    '1.2.840.113556.1.2.1', # instanceType
  ]
)


class DNWithOctetString(DistinguishedName):
  oid = '1.2.840.113556.1.4.903'
  desc = 'DNWithOctetString'

  def displayValue(self,valueindex=0,commandbutton=0):
    octet_tag,count,octet_string,dn = self.attrValue.split(':')
    return ':'.join([
      octet_tag,count,octet_string,
      DisplayDN(self._sid,self._form,self._ls,dn,commandbutton=commandbutton)
    ])


class DNWithString(DNWithOctetString):
  oid = '1.2.840.113556.1.4.904'
  desc = 'DNWithString'


class MicrosoftLargeInteger(Integer):
  oid = '1.2.840.113556.1.4.906'
  desc = 'Integer guaranteed to support 64 bit numbers'


class ObjectSecurityDescriptor(OctetString):
  oid = '1.2.840.113556.1.4.907'
  desc = 'Object-Security-Descriptor'


class Interval(MicrosoftLargeInteger):
  oid = 'Interval-oid'
  desc = 'Large integer with timestamp expressed as 100 nanoseconds since 1601-01-01 00:00'

  def displayValue(self,valueindex=0,commandbutton=0):
    if self.attrValue=='9223372036854775807':
      return '-1: unlimited/off'
    else:
      delta = (long(self.attrValue)-116444736000000000L)/10000000
      if delta>=0:
        return utctime.strftimeiso8601(time.gmtime(delta)).encode('ascii')
      else:
        return self.attrValue.encode('ascii')


class DomainControllerFunctionality(SelectList):
  oid = 'DomainControllerFunctionality-oid'
  desc = 'Functional level of domain controller'

  attr_value_dict = {
    '':'',
    '0':'Windows 2000 Mode',
    '2':'Windows Server 2003 Mode',
    '3':'Windows Server 2008 Mode',
  }

syntax_registry.registerAttrType(
  DomainControllerFunctionality.oid,[
    'domainControllerFunctionality', # no schema information available
  ]
)


class DomainFunctionality(SelectList):
  oid = 'DomainFunctionality-oid'
  desc = 'Functional level of domain'

  attr_value_dict = {
    '':'',
    '0':'Windows 2000 Mixed',
    '1':'Windows 2000 Native',
    '2':'Windows 2003 Interim',
    '3':'Windows 2003',
  }

syntax_registry.registerAttrType(
  DomainFunctionality.oid,[
    'domainFunctionality', # no schema information available
  ]
)


class ForestFunctionality(SelectList):
  oid = 'ForestFunctionality-oid'
  desc = 'Functional level of forest'

  attr_value_dict = {
    '':'',
    '0':'Windows 2000',
    '1':'Windows 2003 Interim',
    '2':'Windows 2003',
  }

syntax_registry.registerAttrType(
  ForestFunctionality.oid,[
    'forestFunctionality', # no schema information available
  ]
)


# Register certain attribute types for Interval
syntax_registry.registerAttrType(
  Interval.oid,[
    '1.2.840.113556.1.4.159',  # accountExpires
    '1.2.840.113556.1.4.49',   # badPasswordTime
    '1.2.840.113556.1.4.52',   # lastLogon
    '1.2.840.113556.1.4.1696', # lastLogonTimestamp
    '1.2.840.113556.1.4.51',   # lastLogoff
    '1.2.840.113556.1.4.96',   # pwdLastSet
  ]
)


class ServerStatus(SelectList):
  oid = 'ServerStatus-oid'
  desc = 'Specifies whether the server is enabled or disabled.'
  attr_value_dict = {
    '':'',
    '1':'enabled',
    '2':'disabled',
  }

syntax_registry.registerAttrType(
  ServerStatus.oid,[
    '1.2.840.113556.1.4.154', # serverStatus
  ]
)


class PwdProperties(BitArrayInteger):
  """
  http://msdn.microsoft.com/en-us/library/ms679431(VS.85).aspx
  """
  oid = 'PwdProperties-oid'
  flag_desc_table = (
    ('DOMAIN_PASSWORD_COMPLEX',1),
    ('DOMAIN_PASSWORD_NO_ANON_CHANGE',2),
    ('DOMAIN_PASSWORD_NO_CLEAR_CHANGE',4),
    ('DOMAIN_LOCKOUT_ADMINS',8),
    ('DOMAIN_PASSWORD_STORE_CLEARTEXT',16),
    ('DOMAIN_REFUSE_PASSWORD_CHANGE',32)
  )

syntax_registry.registerAttrType(
  PwdProperties.oid,[
    '1.2.840.113556.1.4.93', # pwdProperties
  ]
)


# Register all syntax classes in this module
for symbol_name in dir():
  syntax_registry.registerSyntaxClass(eval(symbol_name))



# Work around for Active Directory of Windows 2000:
# Register syntaxes by odd names
syntax_registry.oid2syntax['Boolean'] = Boolean
syntax_registry.oid2syntax['DN'] = DistinguishedName
syntax_registry.oid2syntax['Integer'] = Integer
syntax_registry.oid2syntax['DirectoryString'] = DirectoryString
syntax_registry.oid2syntax['GeneralizedTime'] = GeneralizedTime

syntax_registry.registerAttrType(
  DistinguishedName.oid,[
    'configurationNamingContext',
    'defaultNamingContext',
    'dsServiceName',
    'rootDomainNamingContext',
    'schemaNamingContext',
  ]
)

# MS AD declares these attributes with OctetString
# syntax but Binary syntax is more suitable
syntax_registry.registerAttrType(
  Binary.oid,[
    '1.2.840.113556.1.4.645',    # userCert
  ]
)

# MS AD declares these attributes with DirectoryString
# syntax but OctetString syntax is more suitable
syntax_registry.registerAttrType(
  OctetString.oid,[
    '1.2.840.113556.1.4.138',    # userParameters
  ]
)

syntax_registry.registerAttrType(
  Uri.oid,[
    '1.2.840.113556.1.4.583', # meetingURL
    '1.2.840.113556.1.2.464', # wWWHomePage
    '1.2.840.113556.1.4.749', # url
  ]
)

